module Exam.Exercise1.SemEq where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym; cong; cong₂)
open import Data.Product                          using (_×_; proj₁; proj₂; ∃; ∃-syntax) renaming (_,_ to ⟨_,_⟩)

open import Exam.Exercise1.Syntax
open import Exam.Exercise1.Typing
open import Exam.Exercise1.SemanticsSmallStep renaming (Value to IsValue)
open import Exam.Exercise1.SemanticsBigStep
open import Exam.Exercise1.SoundnessSmallStep using (preserve)

-- Big-step → Small-step -------------------------------------------------------

-- Lemmas you might find useful:

-- A big-step value is also a small-step value.
Value→IsValue : ∀ v →
  IsValue ⌞ v ⌟
Value→IsValue = {!!}

-- You may need multiple lemmas like:
ξ-·₁' :
  M —↠ M′ →
  -------------------
  (M · N) —↠ (M · N′)
ξ-·₁' = {!!}

-- Note that ξ-·₂' is like the ξ-·₂ constructor but with —↠ instead of —→.

-- The theorem you need to prove:

to—↠ : ∀ {M N} →
  M ⇓ N →
  M —↠ ⌞ N ⌟
to—↠ = {!!}

-- Small-step → Big-step -------------------------------------------------------

-- Lemmas you might find useful:

-- Values evaluate to themselves.
⇓-refl-val : ∀ V →
  ⌞ V ⌟ ⇓ V
⇓-refl-val = {!!}

-- A small-step value is also a big-step value.
IsValue→Value : ∀ {M} →
  IsValue M →
  ∃[ V ] (M ≡ ⌞ V ⌟)
IsValue→Value = {!!}

-- The —↠ relation, which we defined in SemanticsSmallStep.agda, is problematic
-- for the proof of our main theorem to⇓:
--
--   to⇓ : ∀ {M V} →
--     M —↠ ⌞ V ⌟ →
--     M ⇓ V
--
-- If you try to prove the main theorem, you will get stuck in the case of the
-- transitivity rule (try it out). The problem is that the transitivity rule,
-- splits the `M —↠ ⌞ V ⌟` into two parts M —↠ N and N —↠ ⌞ V ⌟. To the second
-- part we can easily apply our induction hypothesis to⇓ which gives us L ⇓ V,
-- but with the first part we can't do anything: N does not need to be a value,
-- which means that M —↠ N can be a partial evaluation, which we can't express
-- with the big-step semantics, which only describes the complete evaluation of
-- a program, down to it's value.
--
-- An easy way around this, is to give an alternative definition of the the —↠
-- relation, which we call ↠ and which does not have this transitivity rule, and
-- then prove that both relations imply each other. This allows us to first prove
--
--   to⇓' : ∀ {M V} →
--     M ↠ ⌞ V ⌟ →         -- Note the arrow is a short ↠ instead of —↠
--     M ⇓ V
--
-- and then prove to⇓ by using to⇓' and the fact that —↠ and ↠ are equivalent.

infixr 7 _↠_
data _↠_ : Term → Term → Set where
  ↠-refl :
    M ↠ M

  ↠-step :
    L —→ M →      -- Note the single step —→ in contrast to the multi step —↠
    M ↠ N →       -- used in —↠-trans.
    ----------
    L ↠ N

-- Allthough the ↠ relation does not have transitivity as a constructor,
-- it still enjoy transitivity as a property. This allows us to use
-- transitivity if we need it, but we don't need to cover it as a case
-- when doing induction or pattern matching on a M ↠ N proof, since
-- it's derived from the other constructors.
↠-trans :
  L ↠ M →
  M ↠ N →
  L ↠ N
↠-trans = {!!}

-- The next two lemmas show that ↠ and —↠ imply each other.

↠-to-—↠ :
  M ↠ N →
  M —↠ N
↠-to-—↠ = {!!}

—↠-to-↠ :
  M —↠ N →
  M ↠ N
—↠-to-↠ = {!!}

-- The next lemma is the workhorse for the to⇓' proof.
-- It allows us to extend a big-step proof N ⇓ V with a small-step proof M —→ N.

ext-⇓-with-—→ : ∀ {M N V} →
  M —→ N →
  N ⇓ V →
  M ⇓ V
ext-⇓-with-—→ = {!!}

-- The to⇓' lemma was explained above.

to⇓' : ∀ {M N} →
  M ↠ ⌞ N ⌟ →
  M ⇓ N
to⇓' = {!!}

-- The theorem you need to prove:

to⇓ : ∀ {M N} →
  M —↠ ⌞ N ⌟ →
  M ⇓ N
to⇓ = {!!}

-- Big-step: Alternative Soundness Proof ---------------------------------------

-- Now that we know that the small-step semantics is sound and that the
-- small-step and big-step semantics are equivalent, there is also an easier
-- way to prove type soundness of the big-step semantics:
--
-- If we know term M has type T, and M ⇓ V, then by the semantic equivalence we
-- know that M —↠ ⌞ V ⌟, so by soundness of the small step semantics, we know
-- that ⌞ V ⌟ also has type T.
--
-- Give an alternative type soundness proof for the big-step semantics following
-- the reasoning of the previous paragraph.
--
-- (It should go without saying, that this proof does not count as a solution
-- for the original soundness proof of the previous exercise part ;-)

soundness' : ∀ {M : Term} {T : Type} {V : Value} →
  ∅ ⊢ M ⦂ T →
  M ⇓ V →
  ∅ ⊢ ⌞ V ⌟ ⦂ T
soundness' = {!!}
