module Exam.Exercise1.SemanticsSmallStep where

open import Relation.Nullary using (Dec; yes; no; ¬_)
open import Data.String      using (_≟_)

open import Exam.Exercise1.Syntax
open import Exam.Exercise1.Substitution

-- Small-Step Semantics --------------------------------------------------------

data Value : Term → Set where

  V-ƛ :
    Value (ƛ x ⇒ M)

  V-zero :
    Value `zero

  V-suc :
    Value M →
    ----------
    Value (`suc M)

-- single-step reduction
infix 4 _—→_

data _—→_ : Term → Term → Set where

  β-ƛ :                         -- beta value reduction
                                -- left side: (beta) redex
    Value N →                   -- right side: reductum (generally)
    ------------------------------
    (ƛ x ⇒ M) · N —→ M [ x := N ]

  β-zero :

    case `zero [zero⇒ M |suc x ⇒ N ] —→ M

  β-suc :

    Value L →
    --------------------------------------------------
    case (`suc L) [zero⇒ M |suc x ⇒ N ] —→ N [ x := L ]

  β-μ :

    μ x ⇒ M —→ M [ x := μ x ⇒ M ]

  -- context rules

  ξ-·₁ :

    M —→ M′ →
    --------------------
    M · N —→ M′ · N

  ξ-·₂ :

    Value M →
    N —→ N′ →
    --------------------
    M · N —→ M · N′

  ξ-suc :

    M —→ M′ →
    --------------------
    `suc M —→ `suc M′

  ξ-case :

    L —→ L′ →
    ---------------------------------------------------------------
    case L [zero⇒ M |suc x ⇒ N ] —→ case L′ [zero⇒ M |suc x ⇒ N ]

-- reflexive transitive closure

data _—↠_ : Term → Term → Set where

  —↠-refl :
    M —↠ M

  —↠-step :
    M —→ N →
    ----------
    M —↠ N

  —↠-trans :
    M —↠ N →
    N —↠ L →
    ----------
    M —↠ L

infixr 7 _—↠_
