module Exam.Exercise2.Syntax where

open import Data.String using (String)
open import Data.Nat    using (ℕ)

Id : Set
Id = String

-- A location (pointer/reference) is simply represented as a natural number.
-- We use the letter ℓ to refer to locations (\ell in LaTeX).
Loc : Set
Loc = ℕ

data Term : Set where
  `_                   : Id → Term
  ƛ_⇒_                 : Id → Term → Term
  _·_                  : Term → Term → Term
  `zero                : Term
  `suc_                : Term → Term
  case_[zero⇒_|suc_⇒_] : Term → Term → Id → Term → Term
  μ_⇒_                 : Id → Term → Term

  -- The following constructors are new:

  `new                 : Term → Term         -- Create a new memory cell with a certain value and return its location.
  !_                   : Term → Term         -- Retrieve the value stored at a location.
  _≔_                  : Term → Term → Term  -- Store a new value at a location.
  `loc                 : Loc → Term          -- A location value (not presented to the programmer)
  `tt                  : Term                -- The unit value (returned by evaluating _≔_, to express the abscence of a result)

-- The `loc term corresponds to an actual pointer value.
-- In most languages the `loc term is not exposed to the programmer directly, as
-- this would corresponds to defining a pointer by explicitly giving the memory
-- address.
-- Instead, the usual way to create a `loc term, is as a result of evaluating a `new term.
-- This ensures that the `loc term actually refers to a valid memory location.

infix  5  ƛ_⇒_  μ_⇒_
infix  6  _≔_
infixr 7  !_
infixl 7  _·_
infix  8  `suc_
infix  9  `_

variable
  x y : Id
  L M N L′ M′ N′ : Term
  ℓ ℓ₁ ℓ₂ ℓ' : Loc

-- To reconstruct the example from the pdf, our language is still missing the
-- `let_=_in_` term, but luckily we can simply simulate this term using
-- a lambda-function directly followed by a function application:

infixr 4 `let_≡_`in_
`let_≡_`in_ : Id → Term → Term → Term
`let x ≡ M `in N = (ƛ x ⇒ N) · M

-- It is easy to verify that this definition behaves as expected: the term
--
--   `let x ≡ `zero `in
--   (add · x · x)
--
-- is by definition the same as
--
--   (ƛ x ⇒ add · x · x) · `zero
--
-- which reduces to
--
--   add · `zero · `zero

-- We can now write the example as followed:

example-term : Term
example-term =
  `let "r" ≡ `new `zero `in
  `let "x" ≡ ! (` "r") `in                -- x will be `zero
  `let "_" ≡ (` "r" ≔ `suc `zero) `in
  `let "y" ≡ ! (` "r") `in                -- y will be (`suc `zero)
  `tt
