module Exam.Exercise2.Typing where

open import Data.List                             using (List; _∷_; []; length)
open import Data.Maybe                            using (Maybe; just; nothing)
open import Relation.Binary.PropositionalEquality using (_≡_; _≢_)

open import Exam.Exercise2.Syntax
open import Exam.Exercise2.Sublist
open import Exam.Exercise2.SemanticsSmallStep     using (Store)
open import Exam.Exercise2.All2

-- Type Syntax -----------------------------------------------------------------

data Type : Set where
  _⇒_  : Type → Type → Type
  `ℕ   : Type

  -- New types:

  `ref : Type → Type   -- (`ref T) is the type of locations containing values of type T.
  `⊤   : Type          -- `⊤ is the unit type, i.e. the type with the single value `tt.

variable A B : Type

-- Typing Contexts -------------------------------------------------------------

-- Typing contexts are just as before.

infixl 5 _,_⦂_

data Context : Set where
  ∅ : Context
  _,_⦂_ : Context → Id → Type → Context

variable Γ Γ₁ Γ₂ : Context

-- A relation to find a type assumption in the context: Γ ∋ x ⦂ A

infix 4 _∋_⦂_

data _∋_⦂_ : Context → Id → Type → Set where

  -- variable appears at top of context
  Z :
    Γ , x ⦂ A ∋ x ⦂ A

  -- variable appears somewhere deeper down
  S :
    x ≢ y →
    Γ ∋ x ⦂ A →
    --------------------
    Γ , y ⦂ B ∋ x ⦂ A

-- Store Typings ---------------------------------------------------------------

-- Next, we introduce a store typing, which associates to each location a type:

StoreTy = List Type

variable Σ Σ′ Σ₁ Σ₂ Σ₃ : StoreTy

-- The role of this store typing is similar to the role of the context Γ.
-- Just as we use a context Γ to keep track of the types of our variables,
-- we use a store typing Σ to keep track of the types of our locations
-- (or more precisely the types of the values stored at our locations).
-- Without a store typing, we would run into the same problem as for variables:
-- the term (`loc 23) doesn't have a type by itself. It depends on what value
-- is currently stored at location 23.

-- Hence, our typing relation now has the form (Σ ; Γ ⊢ M ⦂ A), which reads as:
--   "If our locations have types from Σ and our variables have types from Γ,
--    then the term M has type A"

infix 4 _;_⊢_⦂_

data _;_⊢_⦂_ : StoreTy → Context → Term → Type → Set where

  -- The old typing rules basically stay the same, they just
  -- propagate the Σ to subterms, in case one of the subterms
  -- contains a location.

  var :
    Γ ∋ x ⦂ A →
    -----------
    Σ ; Γ ⊢ ` x ⦂ A

  -- rules for functions

  ⇒-I :
    Σ ; Γ , x ⦂ A ⊢ M ⦂ B →
    --------------------
    Σ ; Γ ⊢ ƛ x ⇒ M ⦂ A ⇒ B

  ⇒-E :
    Σ ; Γ ⊢ M ⦂ A ⇒ B →
    Σ ; Γ ⊢ N ⦂ A →
    ----------------
    Σ ; Γ ⊢ M · N ⦂ B

  -- rules for terms denoting numbers

  ℕ-I-zero :
    Σ ; Γ ⊢ `zero ⦂ `ℕ

  ℕ-I-suc :
    Σ ; Γ ⊢ M ⦂ `ℕ →
    --------------------
    Σ ; Γ ⊢ `suc M ⦂ `ℕ

  ℕ-E-case :
    Σ ; Γ ⊢ L ⦂ `ℕ →
    Σ ; Γ ⊢ M ⦂ A →
    Σ ; Γ , x ⦂ `ℕ ⊢ N ⦂ A →
    ----------------------------------------
    Σ ; Γ ⊢ case L [zero⇒ M |suc x ⇒ N ] ⦂ A

  μ-E :
    Σ ; Γ , x ⦂ A ⊢ M ⦂ A →
    --------------------
    Σ ; Γ ⊢ μ x ⇒ M ⦂ A

  -- The new typing rules look as followed:

  ref-I-new :
    Σ ; Γ ⊢ M ⦂ A →           -- If M has type A, then (`new M) is going to give
    -------------------       -- us a reference to a value of type A.
    Σ ; Γ ⊢ `new M ⦂ `ref A


  ref-I-loc :                -- A location has type (`ref A), if the store typing
    lookup ℓ Σ ≡ just A →    -- says that the location stores a value of type A.
    -------------------
    Σ ; Γ ⊢ `loc ℓ ⦂ `ref A

  ref-E-get :                -- If M is a reference to a value of type A, then
    Σ ; Γ ⊢ M ⦂ `ref A →     -- (! M) is supposed to retrieve the value behind
    ----------------         -- the reference, so (! M) has type A.
    Σ ; Γ ⊢ ! M ⦂ A

  ref-E-set :               -- If M is a reference to a value of type A,
    Σ ; Γ ⊢ M ⦂ `ref A →    -- and N has type A, then (M ≔ N) is supposed to
    Σ ; Γ ⊢ N ⦂ A →         -- update the value behind M with the value of N.
    --------------------
    Σ ; Γ ⊢ M ≔ N ⦂ `⊤

  ⊤-I :                    -- The unit term `tt has the unit type `⊤.
    Σ ; Γ ⊢ `tt ⦂ `⊤
